@@ -1,8 +1,8 @@
-import os
 import traceback
 from enum import Enum
+from ostium_python_sdk.constants import PRECISION_2
 from web3 import Web3
-from .abi import usdc_abi, ostium_trading_abi, ostium_trading_storage_abi
+from .abi.abi import usdc_abi, ostium_trading_abi, ostium_trading_storage_abi
 from .utils import convert_to_scaled_integer, fromErrorCodeToMessage, get_tp_sl_prices, to_base_units
 from eth_account.account import Account
 
@@ -14,12 +14,9 @@ class OpenOrderType(Enum):
 
 
 class Ostium:
-    def __init__(self, w3: Web3, usdc_address: str, ostium_trading_storage_address: str, ostium_trading_address: str, private_key: str) -> None:
-        if not private_key:
-            raise ValueError(
-                "Private key is required for Ostium initialization")
-
+    def __init__(self, w3: Web3, usdc_address: str, ostium_trading_storage_address: str, ostium_trading_address: str, private_key: str, verbose=False) -> None:
         self.web3 = w3
+        self.verbose = verbose
         self.private_key = private_key
         self.usdc_address = usdc_address
         self.ostium_trading_storage_address = ostium_trading_storage_address
@@ -33,7 +30,24 @@ class Ostium:
         self.ostium_trading_contract = self.web3.eth.contract(
             address=self.ostium_trading_address, abi=ostium_trading_abi)
 
+        self.slippage_percentage = 2  # 2%
+
+    def log(self, message):
+        if self.verbose:
+            print(message)
+
+    def set_slippage_percentage(self, slippage_percentage):
+        self.slippage_percentage = slippage_percentage
+
+    def get_slippage_percentage(self):
+        return self.slippage_percentage
+
+    def get_public_address(self):
+        public_address = self._get_account().address
+        return public_address
+
     def _get_account(self) -> Account:
+        self._check_private_key()
         """Get account from stored private key"""
         return self.web3.eth.account.from_key(self.private_key)
 
@@ -43,13 +57,19 @@ class Ostium:
     def get_nonce(self, address):
         return self.web3.eth.get_transaction_count(address)
 
+    def _check_private_key(self):
+        if not self.private_key:
+            raise ValueError(
+                "Private key is required for Ostium platform write-operations")
+
     def perform_trade(self, trade_params, at_price):
+        self.log(f"Performing trade with params: {trade_params}")
         account = self._get_account()
         amount = to_base_units(trade_params['collateral'], decimals=6)
         self.__approve(account, amount)
 
         try:
-            print("Final trade parameters being sent:", trade_params)
+            self.log(f"Final trade parameters being sent: {trade_params}")
             tp_price, sl_price = get_tp_sl_prices(trade_params)
 
             trade = {
@@ -77,7 +97,7 @@ class Ostium:
                     raise Exception('Invalid order type')
 
             trade_tx = self.ostium_trading_contract.functions.openTrade(
-                trade, order_type, 9000).build_transaction({'from': account.address})
+                trade, order_type, int(self.slippage_percentage * PRECISION_2)).build_transaction({'from': account.address})
             trade_tx['nonce'] = self.get_nonce(account.address)
 
             signed_tx = self.web3.eth.account.sign_transaction(
@@ -86,55 +106,80 @@ class Ostium:
                 signed_tx.raw_transaction)
             trade_receipt = self.web3.eth.wait_for_transaction_receipt(
                 trade_tx_hash)
-            print('Trade Receipt:', trade_receipt)
+            self.log(f"Trade Receipt: {trade_receipt}")
             return trade_receipt
 
         except Exception as e:
-            reason_string, suggestion = fromErrorCodeToMessage(e)
+            reason_string, suggestion = fromErrorCodeToMessage(
+                e, verbose=self.verbose)
             print(
                 f"An error ({str(e)}) occurred during the trading process - parsed as {reason_string}")
             raise Exception(
                 f'{reason_string}\n\n{suggestion}' if suggestion != None else reason_string)
 
-    def cancel_limit_order(self, pairID, index):
+    def cancel_limit_order(self, pair_id, trade_index):
         account = self._get_account()
 
         trade_tx = self.ostium_trading_contract.functions.cancelOpenLimitOrder(
-            int(pairID), int(index)).build_transaction({'from': account.address})
+            int(pair_id), int(trade_index)).build_transaction({'from': account.address})
         trade_tx['nonce'] = self.get_nonce(account.address)
 
         signed_tx = self.web3.eth.account.sign_transaction(
             trade_tx, private_key=self.private_key)
         trade_tx_hash = self.web3.eth.send_raw_transaction(
             signed_tx.raw_transaction)
-        print('Cancel Limit Order TX Hash:', trade_tx_hash.hex())
+        self.log(f"Cancel Limit Order TX Hash: {trade_tx_hash.hex()}")
 
         trade_receipt = self.web3.eth.wait_for_transaction_receipt(
             trade_tx_hash)
-        print('Cancel Limit Order Receipt:', trade_receipt)
+        self.log(f"Cancel Limit Order Receipt: {trade_receipt}")
         return trade_receipt
 
-    def close_trade(self, pairID, index):
+    def close_trade(self, pair_id, trade_index, close_percentage = 100):
+        self.log(f"Closing trade for pair {pair_id}, index {trade_index}")
         account = self._get_account()
 
+        close_percentage = to_base_units(close_percentage, decimals=2)
+
         trade_tx = self.ostium_trading_contract.functions.closeTradeMarket(
-            int(pairID), int(index)).build_transaction({'from': account.address})
+            int(pair_id), int(trade_index), int(close_percentage)).build_transaction({'from': account.address})
         trade_tx['nonce'] = self.get_nonce(account.address)
 
         signed_tx = self.web3.eth.account.sign_transaction(
             trade_tx, private_key=self.private_key)
         trade_tx_hash = self.web3.eth.send_raw_transaction(
             signed_tx.raw_transaction)
-        print('Trade TX Hash:', trade_tx_hash.hex())
+        self.log(f"Trade TX Hash: {trade_tx_hash.hex()}")
 
         trade_receipt = self.web3.eth.wait_for_transaction_receipt(
             trade_tx_hash)
-        print('Trade Receipt:', trade_receipt)
+        self.log(f"Trade Receipt: {trade_receipt}")
         return trade_receipt
 
+    def remove_collateral(self, pair_id, trade_index, remove_amount):
+        self.log(f"Remove collateral for trade for pair {pair_id}, index {trade_index}: {remove_amount} USDC")
+        account = self._get_account()
+
+        amount = to_base_units(remove_amount, decimals=6)
+
+        trade_tx = self.ostium_trading_contract.functions.removeCollateral(
+            int(pair_id), int(trade_index), int(amount)).build_transaction({'from': account.address})
+        trade_tx['nonce'] = self.get_nonce(account.address)
+
+        signed_tx = self.web3.eth.account.sign_transaction(
+            trade_tx, private_key=self.private_key)
+        trade_tx_hash = self.web3.eth.send_raw_transaction(
+            signed_tx.raw_transaction)
+        self.log(f"Remove Collateral TX Hash: {trade_tx_hash.hex()}")
+
+        remove_receipt = self.web3.eth.wait_for_transaction_receipt(
+            trade_tx_hash)
+        self.log(f"Remove Collateral Receipt: {remove_receipt}")
+        return remove_receipt
+
     def add_collateral(self, pairID, index, collateral):
+        account = self._get_account()
         try:
-            account = self._get_account()
             amount = to_base_units(collateral, decimals=6)
             self.__approve(account, amount)
 
@@ -146,11 +191,11 @@ class Ostium:
                 add_collateral_tx, private_key=self.private_key)
             add_collateral_tx_hash = self.web3.eth.send_raw_transaction(
                 signed_tx.raw_transaction)
-            print('Add Collateral TX Hash:', add_collateral_tx_hash.hex())
+            self.log(f"Add Collateral TX Hash: {add_collateral_tx_hash.hex()}")
 
             add_collateral_receipt = self.web3.eth.wait_for_transaction_receipt(
                 add_collateral_tx_hash)
-            print('Add Collateral Receipt:', add_collateral_receipt)
+            self.log(f"Add Collateral Receipt: {add_collateral_receipt}")
             return add_collateral_receipt
 
         except Exception as e:
@@ -158,20 +203,22 @@ class Ostium:
             traceback.print_exc()
             raise e
 
-    def update_tp(self, pairID, index, tp):
+    def update_tp(self, pair_id, trade_index, tp_price):
+        self.log(
+            f"Updating TP for pair {pair_id}, index {trade_index} to {tp_price}")
+        account = self._get_account()
         try:
-            account = self._get_account()
-            tp_value = to_base_units(tp, decimals=18)
+            tp_value = to_base_units(tp_price, decimals=18)
 
             update_tp_tx = self.ostium_trading_contract.functions.updateTp(
-                int(pairID), int(index), tp_value).build_transaction({'from': account.address})
+                int(pair_id), int(trade_index), tp_value).build_transaction({'from': account.address})
             update_tp_tx['nonce'] = self.get_nonce(account.address)
 
             signed_tx = self.web3.eth.account.sign_transaction(
                 update_tp_tx, private_key=self.private_key)
             update_tp_tx_hash = self.web3.eth.send_raw_transaction(
                 signed_tx.raw_transaction)
-            print('Update TP TX Hash:', update_tp_tx_hash.hex())
+            self.log(f"Update TP TX Hash: {update_tp_tx_hash.hex()}")
 
         except Exception as e:
             print("An error occurred during the update tp process:")
@@ -179,8 +226,8 @@ class Ostium:
             raise e
 
     def update_sl(self, pairID, index, sl):
+        account = self._get_account()
         try:
-            account = self._get_account()
             sl_value = to_base_units(sl, decimals=18)
 
             update_sl_tx = self.ostium_trading_contract.functions.updateSl(
@@ -191,10 +238,11 @@ class Ostium:
                 update_sl_tx, private_key=self.private_key)
             update_sl_tx_hash = self.web3.eth.send_raw_transaction(
                 signed_tx.raw_transaction)
-            print('Update SL TX Hash:', update_sl_tx_hash.hex())
+            self.log(f"Update SL TX Hash: {update_sl_tx_hash.hex()}")
 
         except Exception as e:
-            reason_string, suggestion = fromErrorCodeToMessage(str(e))
+            reason_string, suggestion = fromErrorCodeToMessage(
+                str(e), verbose=self.verbose)
             print(
                 f"An error occurred during the update sl process: {reason_string}")
             raise Exception(
@@ -216,15 +264,16 @@ class Ostium:
                 approve_tx, private_key=self.private_key)
             approve_tx_hash = self.web3.eth.send_raw_transaction(
                 signed_tx.raw_transaction)
-            print('Approval TX Hash:', approve_tx_hash.hex())
+            self.log(f"Approval TX Hash: {approve_tx_hash.hex()}")
 
             approve_receipt = self.web3.eth.wait_for_transaction_receipt(
                 approve_tx_hash)
-            print('Approval Receipt:', approve_receipt)
+            self.log(f"Approval Receipt: {approve_receipt}")
 
     def withdraw(self, amount, receiving_address):
+        account = self._get_account()
+
         try:
-            account = self._get_account()
             amount_in_base_units = to_base_units(amount, decimals=6)
 
             if not self.web3.is_address(receiving_address):
@@ -241,15 +290,16 @@ class Ostium:
                 transfer_tx, private_key=self.private_key)
             transfer_tx_hash = self.web3.eth.send_raw_transaction(
                 signed_tx.raw_transaction)
-            print('Transfer TX Hash:', transfer_tx_hash.hex())
+            self.log(f"Transfer TX Hash: {transfer_tx_hash.hex()}")
 
             transfer_receipt = self.web3.eth.wait_for_transaction_receipt(
                 transfer_tx_hash)
-            print('Transfer Receipt:', transfer_receipt)
+            self.log(f"Transfer Receipt: {transfer_receipt}")
             return transfer_receipt
 
         except Exception as e:
-            reason_string, suggestion = fromErrorCodeToMessage(str(e))
+            reason_string, suggestion = fromErrorCodeToMessage(
+                str(e), verbose=self.verbose)
             print(
                 f"An error occurred during the transfer process: {reason_string}")
             raise Exception(
@@ -257,10 +307,7 @@ class Ostium:
 
     def update_limit_order(self, pair_id, index, pvt_key, price=None, tp=None, sl=None):
         try:
-            print('update_limit_order called with pair_id', pair_id, 'index',
-                  index, 'pvt_key', pvt_key, 'price', price, 'tp', tp, 'sl', sl)
             account = self.web3.eth.account.from_key(pvt_key)
-
             # Get existing order details (tbd why read from storage)
             existing_order = self.ostium_trading_storage_contract.functions.getOpenLimitOrder(
                 account.address,
@@ -268,7 +315,7 @@ class Ostium:
                 int(index)
             ).call()
 
-            print('existing_order', existing_order)
+            self.log(f"existing_order {existing_order}")
             # Use existing values if new values are not provided
             price_value = convert_to_scaled_integer(
                 price) if price is not None else existing_order[1]  # openPrice
@@ -277,8 +324,6 @@ class Ostium:
             sl_value = convert_to_scaled_integer(
                 sl) if sl is not None else existing_order[3]    # sl
 
-            print('calling updateOpenLimitOrder with price_value',
-                  price_value, 'tp_value', tp_value, 'sl_value', sl_value)
             trade_tx = self.ostium_trading_contract.functions.updateOpenLimitOrder(
                 int(pair_id),
                 int(index),
@@ -293,15 +338,16 @@ class Ostium:
                 trade_tx, private_key=account.key)
             trade_tx_hash = self.web3.eth.send_raw_transaction(
                 signed_tx.raw_transaction)
-            print('Update Limit Order TX Hash:', trade_tx_hash.hex())
+            self.log(f"Update Limit Order TX Hash: {trade_tx_hash.hex()}")
 
             trade_receipt = self.web3.eth.wait_for_transaction_receipt(
                 trade_tx_hash)
-            print('Update Limit Order Receipt:', trade_receipt)
+            self.log(f"Update Limit Order Receipt: {trade_receipt}")
             return trade_receipt
 
         except Exception as e:
-            reason_string, suggestion = fromErrorCodeToMessage(str(e))
+            reason_string, suggestion = fromErrorCodeToMessage(
+                str(e), verbose=self.verbose)
             print(
                 f"An error occurred during the update limit order process: {reason_string}")
             raise Exception(
